/* -*- mode:prolog; -*-
 ******************************************************************
 * $Id: forest.pl,v 1.6 2004/02/19 09:30:46 clerger Exp $
 * Copyright (C) 1999, 2004 by INRIA 
 * Author: Eric de la Clergerie <Eric.De_La_Clergerie@inria.fr>
 * ----------------------------------------------------------------
 *
 *  forest.pl -- Forest Brwoser
 *
 * ----------------------------------------------------------------
 * Description
 * 
 * ----------------------------------------------------------------
 */


:-module(forest).

:-import{ file=>'format.pl', preds => [format/2] }.

:-std_prolog update_counter/2,value_counter/2.
:-std_prolog display_forest/1.

:-std_prolog forest/2.

:-extensional forest_type/2.

:-std_prolog wrapped_forest/1.

:-finite_set(forest,[init,call,and,or,join,indirect,rand,trace]).

:-end_require.

forest_type(0,init). 		% for initial objects              bptr = init
forest_type(1,call).            % for objects resulting from call  bptr = call
forest_type(2,and).             % for std objects                  bptr = and(trans,item)
forest_type(3,or).              % for alternatives                 bptr = or(bptr1,bptr2)
forest_type(4,join).            % not used
forest_type(5,indirect).        % for inlined objects             bptr = indirect(item) 
forest_type(6,rand).            % not used (reverse and)
forest_type(7,trace).           % to label objects                 bptr = trace(label,bptr1)

%% The indirect type is used for inligning the forest attached to an object
%% rather than introducing a new entry (clause) in the forest
%% It is used for kleene operator

%% The trace type is strangely implemented. The label is actually not
%% related to the trans but to next object to be joined with trans
%% i.e.   and(trans,item) with bptr(trans) = trace(item_label,bptr1)

forest(Addr,Forest) :-
	forest(Addr,Type,AX,AY,L),
	forest_type(Type,FType),
	forest_decode(FType,AX,AY,L,Forest)
	.

:-rec_prolog forest_decode/5.

forest_decode(X::forest[init,call],_,_,_,X).
forest_decode(and,A,B,L,and(A,BB)) :- (L=[] -> BB=B ; BB = (L:B)).
forest_decode(indirect,A,B,_,indirect(A,B)).

update_counter(Name,V) :- 
    recorded( counter(Name,V) ) ->
    ( erase( counter(Name,V) ),W is V+1,record( counter(Name,W) ) );
    ( V=0, record( counter(Name,1) ) ).

value_counter(Name,V) :- (recorded( counter(Name,V) ) -> true ; V = 0).

wrapped_forest( X ) :- 
	(   item_term(A,X),
	    recorded(A,Add_A),
	    build_forest(Add_A,_,_),
	    fail
	;   
	    loop(0)
	)
    .

:-std_prolog loop/1.

loop(Id) :- 
	display_forest(Id),
	New_Id is Id + 1,
	loop(New_Id).

:-light_tabular
	build_forest_aux/3,
	build_forest/3
	.

build_forest_aux(Add,K,L) :-
	forest(Add,Forest),
	%%	format('Forest ~w->~w\n',[Add,Forest]),
	( Forest = forest[call,init] ->
	    K=L
	;   Forest = and(Add_X,Y) ->
	    (	Y = LY:Add_Y xor Y=Add_Y, LY=[] ),
	    (	empty(Add_Y) ->
		build_forest_aux(Add_X,K,L)
	    ;	
		build_forest(Add_Y,Id_Y,_),
		build_forest_aux(Add_X,K,[LY:Id_Y|L])
	    )
	;   
	    format('not yet implemented ~w\n',[Forest])
	).

:-std_prolog empty/1.

empty(Add) :- forest(Add,forest[call,init]).

:-mode(build_forest/3,+(+,-,-)).

build_forest(Add,Id,L) :-
	update_counter(forest_node,Id),
	build_forest_aux(Add,L,[])
	.

display_forest(Id) :-
	'$answers'( build_forest(Add,Id,L) ),
	recorded(A,Add),
	(   item_term(A,X) ->
	    numbervars(X),
	    format('~w\n\t~w <-- ',[X,Id]),
	    every((domain(Label:Id2,L),
		   ( Label = [] ->
		       format( ' ~w', [Id2] )
		   ;   
		       format( ' [~w] ~w', [Label,Id2] )
		   ))),
	    nl
	;
	    true		%% part to skip
	)
	.



